//
// Unit : AsgDebug
//
// Atlantis Software Group * Message Based Debugging
//
// ---------------------------------------------------------------------------------------
//
// Wrapper around several known tools for message based debugging
//
// Write To         Viewer         Source                   Status
// ..........................................................................
// wtIgnore         none           ./.                      internal
// wtCodeSite       CodeSite       www.raize.com            commercial
// wtOverseer       Overseer       delphree.clexpert.com    freeware
// wtGExperts       GEDebug        www.gexperts.com         freeware
// wtDebugger       DbgView        www.sysinternals.com     freeware
// wtConsole        Console        ./.                      internal
// wtFile           Text File      ./.                      internal
// ..........................................................................
//
// ---------------------------------------------------------------------------------------
//
// Copyright (c) 2000 Atlantis Software Group, Franz-Leo Chomse, Uebach-Palenberg, Germany
//
// 2000-01-13 Version 1.00
//
// ---------------------------------------------------------------------------------------
//

unit AsgDebug;

{$DEFINE UseCodeSite}

interface

uses
// Borland Units
  Windows,
  SysUtils,
  Classes,
  Graphics,
  Messages,
  Controls,
  Dialogs,
  TypInfo,
// Algorithms Alfresco (Julian Bucknall)
  AABitSet;

const

// General Messages
// ... Informational message (no attachment)
  dbgmInfo = 101;
// ... Warning message (no attachment)
  dbgmWarning = 102;
// ... Error message (no attachment)
  dbgmError = 103;
// ... Informational messages (highlighted)
  dbgmNote = 104;

// Specific Messages
// ... Automatically numbered check point
  dbgmCheckPoint = 201;
// ... An object's published properties (with current values) is attached
//     is attached to the message
  dbgmObject = 202;
// ... A text-based representation of a component's streaming data is
//     is attached to the message
  dbgmStream = 203;
// ... The current contents of a string list is attached to the message
  dbgmStringList = 204;
// ... The message contains the current value of a property
  dbgmProperty = 205;
// ... Entry point for a method, procedure, or function.
//     The current indent level is incremented by one.
  dbgmEnterMethod = 206;
// ... Exit point for a method, procedure, or function.
//     The current indent level is decremented by one.
  dbgmExitMethod = 207;
// ... a single line separator
  dbgmSeparator = 208;
// ... a wide separate bar.
  dbgmInactiveSeparator = 209;
// ... clear the message list
  dbgmClear = 210;
// ... Directs message to the Scratch Pad pane
  dbgmScratchPad = 211;
// ... The memory dump is attached to the message
  dbgmMemoryDump = 212;

// Maximal number of supported ranges
  dbgRangeLimit = 1024;

type

  TAsgViewerType = (wtIgnore,                               // Ignore all messages
    wtCodeSite,                                             // External Viewer - propriatary messages
    wtOverseer,                                             // External Viewer - propriatary messages
    wtGExperts,                                             // External Viewer - propriatary messages
    wtDebugger,                                             // External Viewer - OutputDebugString
    wtConsole,                                              // Console window
    wtFile);                                                // Text file

  TAsgDebugMessage = class(TPersistent)
  private
// Client
    FViewerType: TAsgViewerType;
    FDateTimeFormat: string;
    FWantEnabled: Boolean;
// Server Simulation
    FIndentLevel: Integer;
    FCheckPointCount: Integer;
    FSeparatorSize: Integer;
    FSimulateAttachments: Boolean;
    FLogFile: TextFile;
    FLogFileActive: Boolean;
    FLogFilePrefix: string;
  protected
    function GetEnabled: Boolean;
    procedure SetEnabled(Value: Boolean);
    procedure SetViewerType(Value: TAsgViewerType);
    procedure SetDateTimeFormat(Value: string);
{$IFDEF UseCodeSite}
    function CodeSiteMsgType(AsgMsgType: Integer): Integer;
{$ENDIF}
    function OverseerMsgType(AsgMsgType: Integer): Integer;
    function GExpertsMsgType(AsgMsgType: Integer): TMsgDlgType;
    function ConsoleMsgType(AsgMsgType: Integer): string;
    procedure InternalClear;
    function ComponentToStringList(Obj: TComponent): TStringList;
    function ObjectToStringList(Obj: TObject): TStringList;
    function MemoryToStringList(var Buffer; BufferSize: Integer): TStringList;
    function PropertyToString(PropOwner: TObject; PropName: string; var PropObject: TObject): string;
    procedure OpenLogFile;
    procedure CloseLogFile;
  public
    constructor Create;
    destructor Destroy;
      override;
    function CheckViewerType(Value: TAsgViewerType): Boolean;
    function CheckEnabled: Boolean;
//  --- Basic Messages
// ... Sends a dbgmInfo message to the Viewer. Use the DbgSendMsgEx method
//     to specify a different message type.
    procedure DbgSendMsg(const Msg: string);
    procedure DbgSendMsgEx(MsgType: Integer; const Msg: string);
// ... Sends a formatted dbgmInfo message to the CodeSite Viewer. The
//     parameters to this message are the same as those used in the
//     standard Format function. Use the DbgSendMsgEx method to specify
//     a different message type.
    procedure DbgSendFmtMsg(const Fmt: string; const Args: array of const);
    procedure DbgSendFmtMsgEx(MsgType: Integer; const Fmt: string; const Args: array of const);
// --- Derived Messages
// ... Sends a dbgmNote message string to the Viewer. To format the message
//     string, use the DbgSendFmtNote method which has the same parameters as
//     the standard Format function.
    procedure DbgSendNote(const Msg: string);
    procedure DbgSendFmtNote(const Fmt: string; const Args: array of const);
// ... Sends a dbgmWarning message string to the Viewer. To format the message
//     string, use the DbgSendFmtNote method which has the same parameters as the
//     standard Format function.
    procedure DbgSendWarning(const Msg: string);
    procedure DbgSendFmtWarning(const Fmt: string; const Args: array of const);
// ... Sends a dbgmError message string to the Viewer. To format the
//     message string, use the DbgSendFmtError method which has the same
//     parameters as the standard Format function.
    procedure DbgSendError(const Msg: string);
    procedure DbgSendFmtError(const Fmt: string; const Args: array of const);
// --- Viewer Control Messages
// ... Sends a dbgmCheckPoint message to the Viewer. The check point index
//     is automatically incremented when the message is sent. Check points
//     are useful when you need to keep track of how many times a particular
//     section of code is being executed. To add just a separator bar to
//     the message list, use the AddSeparator method.
    procedure DbgAddCheckPoint(const CheckName: string);
// ... Sends a dbgmSeparator message to the Viewer. Separators are displayed
//     as thin gray lines in the message list and are used to group messages.
//     To associate an index with a separator, use the AddCheckPoint method.
    procedure DbgAddSeparator;
// ... Instructs the Viewer to clear the message list. When the Clear method
//     is called, the user is not asked to confirm the operation even if the
//     Display clear confirmation dialog option is checked. Note that when
//     the user elects to clear the message list from within the Viewer, the
//     confirmation dialog option is honored.
    procedure DbgClear;
// ... Sends a dbgmEnterMethod message to the Viewer. This instructs the
//     viewer to increment the current indent level, which affects all
//     future messages until a dbgExitMethod message is received or the
//     message list is cleared. Call the DbgEnterMethod method at the begin
//     of a method, procedure, or function to enable Call Stack tracking
//     in the Viewer.
    procedure DbgEnterMethod(const Msg: string);
    procedure DbgEnterMethodEx(const Msg: string; const Value: string);
    procedure DbgEnterMethodFmtEx(const Msg: string; const Fmt: string; const Args: array of const);
// ... Sends a dbgmExitMethod message to the Viewer. This instructs the
//     viewer to decrement the current indent level, which affects all
//     future messages until a dbgmEnterMethod message is received or the
//     message list is cleared. Call the DbgExitMethod method at the end of
//     a method, procedure, or function to enable Call Stack tracking in
//     the Viewer.
    procedure DbgExitMethod(const Msg: string);
    procedure DbgExitMethodEx(const Msg: string; const Value: string);
    procedure DbgExitMethodFmtEx(const Msg: string; const Fmt: string; const Args: array of const);
// ... Sends a non-persistent string message to Scratch Pad pane of the
//     Viewer. The PadIndex parameter is used to specify which scratch pad
//     will receive and display the message. There are 100 scratch pad
//     lines available (0..99). To format the message string, use the
//     DbgFmtScratchPad method, which has the same parameters as the standard
//     Format function.
    procedure DbgScratchPad(LineIndex: Integer; const Msg: string);
    procedure DbgFmtScratchPad(LineIndex: Integer; const Fmt: string; const Args: array of const);
// --- Computed messages
// ... Tests whether the Value parameter is assigned and then passes the
//     result as a dbgmInfo message to the Viewer. Use the DbgSendAssignedEx
//     method to specify a different message type.
    procedure DbgSendAssigned(const Msg: string; Value: Pointer);
    procedure DbgSendAssignedEx(MsgType: Integer; const Msg: string; Value: Pointer);
// ... Evaluates the Expression parameter and then passes the result as
//     a dbgmInfo message to the CodeSite Viewer. Use the DbgSendBooleanEx
//     method to specify a different message type.
    procedure DbgSendBoolean(const Msg: string; Expression: Boolean);
    procedure DbgSendBooleanEx(MsgType: Integer; const Msg: string; Expression: Boolean);
// ... Determines the color symbol name (if available) of the Value parameter
//     and the RGB values for the corresponding color and sends the formatted
//     result as a dbgmInfo message to the Viewer. Use the DbgSendColorEx method
//     to specify a different message type.
    procedure DbgSendColor(const Msg: string; Value: TColor);
    procedure DbgSendColorEx(MsgType: Integer; const Msg: string; Value: TColor);
// ... Formats the Value parameter as a string and then passes the result as
//     a dbgmInfo message to the Viewer. Use the DbgSendIntegerEx method to specify
//     a different message type. To send floating-point numbers to the Viewer,
//     use the DbgSendFloat method.
    procedure DbgSendInteger(const Msg: string; Value: Longint);
    procedure DbgSendIntegerEx(MsgType: Integer; const Msg: string; Value: Longint);
// ... Formats the Value parameter as a string and then passes the result
//     as a dbgmInfo message to the Viewer. Use the DbgSendPointEx method to
//     specify a different message type.
    procedure DbgSendPoint(const Msg: string; Value: TPoint);
    procedure DbgSendPointEx(MsgType: Integer; const Msg: string; Value: TPoint);
// ... Formats the Value parameter as a string and then passes the result as a
//     dbgmInfo message to the Viewer. Use the DbgSendRectEx method to specify a
//     different message type.
    procedure DbgSendRect(const Msg: string; Value: TRect);
    procedure DbgSendRectEx(MsgType: Integer; const Msg: string; Value: TRect);
// ... Formats the Value parameter as a string and then passes the result as
//     a dbgmInfo message to the Viewer. Use the DbgSendStringEx method to specify
//     a different message type.
    procedure DbgSendString(const Msg: string; const Value: string);
    procedure DbgSendStringEx(MsgType: Integer; const Msg: string; const Value: string);
// ... Formats the Value parameter as a string and then passes the result as
//     a dbgmInfo message to the Viewer. Use the DbgSendCurrencyEx method to specify
//     a different message type. The format of the currency value is controlled
//     by the Regional Settings in the Control Panel.
    procedure DbgSendCurrency(const Msg: string; Value: Currency);
    procedure DbgSendCurrencyEx(MsgType: Integer; const Msg: string;
      Value: Currency);
// ... Formats the Value parameter as a string and then passes the result as
//     a dbgmInfo message to the Viewer. Use the DbgSendFloatEx method to specify
//     a different message type. The value is formatted according to the rules
//     of the "n" format string.
    procedure DbgSendFloat(const Msg: string; Value: Extended);
    procedure DbgSendFloatEx(MsgType: Integer; const Msg: string;
      Value: Extended);
// ... Formats the Value parameter as a string and then passes the result as
//     a dbgmInfo message to the Viewer. Use the DbgSendDWordEx method to specify
//     a different message type. Under Delphi 3, DWord maps to a 32-bit signed
//     Integer. Under Delphi 4, DWord maps to a LongWord--a 32-bit unsigned Integer.
    procedure DbgSendDWord(const Msg: string; Value: DWord);
    procedure DbgSendDWordEx(MsgType: Integer; const Msg: string;
      Value: DWord);
// ... Formats the Value parameter as a string and then passes the result as
//     a dbgmInfo message to the Viewer. Use the DbgSendNumAsHex method to specify
//     a different message type. Under Delphi 3, DWord maps to a 32-bit signed
//     Integer. Under Delphi 4, DWord maps to a LongWord--a 32-bit unsigned Integer.
//     Instead of being formatted as a normal base-10 number, the value is formatted
//     as a hexidecimal number. Under Delphi, the hex value is prefixed with
//     a "$", while under C++Builder the hex value is prefixed with a "0x".
    procedure DbgSendNumAsHex(const Msg: string; Value: DWord);
    procedure DbgSendNumAsHexEx(MsgType: Integer; const Msg: string;
      Value: DWord);
// ... Formats the Value parameter as a string and then passes the result as
//     a dbgmInfo message to the Viewer. Use the DbgSendEnumEx method to specify
//     a different message type. The TypeInfo function is used to pass the RTTI for
//     the desired enumeration.
    procedure DbgSendEnum(const Msg: string; EnumTypeInfo: PTypeInfo; Value: Integer);
    procedure DbgSendEnumEx(MsgType: Integer; const Msg: string; EnumTypeInfo: PTypeInfo; Value: Integer);
// ... Formats the Value parameter as a string and then passes the result as
//     a dbgmInfo message to the Viewer. Use the DbgSendDateTimeEx method to
//     specify a different message type. The DateTimeFormat property controls
//     the format used to build the string;
    procedure DbgSendDateTime(const Msg: string; Value: TDateTime);
    procedure DbgSendDateTimeEx(MsgType: Integer; const Msg: string; Value: TDateTime);
// --- Messages with attachments
// ... Sends a dbgmObject message to the CodeSite Viewer. The Msg string is
//     displayed in the message list, while a description of the published
//     properties and their current values for the Obj parameter is displayed
//     in the inspector pane.
    procedure DbgSendObject(const Msg: string; Obj: TObject);
// ... Uses RTTI to determine if a property named PropName exists for Obj.
//     If one does, then the type of the property is determined so that the
//     current value of the property can be correctly displayed. If the
//     property is a class type, then a call to the DbgSendObject method. Otherwise,
//     the current value of the property is formatted along with the Msg string
//     and is sent to the Viewer as a dbgmProperty message.
    procedure DbgSendProperty(const Msg: string; Obj: TObject; const PropName: string);
// ... Sends a dbgmStream message to the Viewer. The Msg string is displayed
//     in the message list, while a description of the data that is transferred
//     when the component referenced in the Obj parameter is placed on a stream
//     is displayed in the inspector pane.
    procedure DbgSendStream(const Msg: string; Obj: TComponent);
// ... Sends a dbgmStringList message to the Viewer. The Msg string is displayed
//     in the message list, while the list of string values currently stored in
//     the List parameter is displayed in the inspector pane.
    procedure DbgSendStringList(const Msg: string; List: TStrings);
    procedure DbgSendStringListEx(MsgType: Integer; const Msg: string; List: TStrings);
// ... Sends a dbgmMemoryDump message to the Viewer. The Msg string is displayed
//     in the message list, while the dump of the given buffer is displayed in
//     the inspector pane.
    procedure DbgSendMemoryDump(const Msg: string; var Buffer; BufferSize: Integer);
  published
// ... When the Enabled property is set to True (by default),
//     the debug message object will send all messages to the
//     Viewer or File.  When the Enabled property is set to
//     False, the messages are ignored.
    property Enabled: Boolean
      read GetEnabled
      write SetEnabled;
// ... Viewer type
    property ViewerType: TAsgViewerType
      read FViewerType
      write SetViewerType;
// ... Date time format
    property DateTimeFormat: string
      read FDateTimeFormat
      write SetDateTimeFormat;
// ... Simulation of attachments
    property SimulateAttachments: Boolean
      read FSimulateAttachments
      write FSimulateAttachments;
// ... Prefix of file mame
    property LogFilePrefix: string
      read FLogFilePrefix
      write FLogFilePrefix;
  end;

  TAsgDebugLevel = class(TPersistent)
  private
    FRangeFlags: TAABitSet;
    FRangeNames: TStringList;
  protected
    function GetRangeCount: Integer;
    function GetRangeFlag(Index: Integer): Boolean;
    procedure SetRangeFlag(Index: Integer; Value: Boolean);
    function GetRangeName(Index: Integer): string;
    procedure SetRangeName(Index: Integer; Value: string);
  public
    constructor Create;
    destructor Destroy;
      override;
// ... Add a new debug range
    function AddRange(const RangeName: string): Integer;
// ... Find the number of a debug range. if the range is not known the
//     Result is -1.
    function FindRange(const RangeName: string): Integer;
    property IsRangeActive[Index: Integer]: Boolean
      read GetRangeFlag
      write SetRangeFlag;
    property RangeName[Index: Integer]: string
      read GetRangeName
      write SetRangeName;
    property RangeCount: integer
      read GetRangeCount;
  end;

var
  DebugMessage: TAsgDebugMessage;
  DebugLevel: TAsgDebuglevel;

implementation

uses
{$IFDEF UseCodeSite}
// Codesite : Raize Software Solutions
  RzCSIntf,
{$ENDIF}
// Overseer : Nexus Software Library
  uDbgIntf,
  uDbg,
// GDebug : GExperts
  dbugintf;

var
  DebugMessageClosing: Boolean;

//
// =======================================================================================
//
// Class Management
//
// =======================================================================================
//

constructor TAsgDebugMessage.Create;
begin
  inherited;
{$IFDEF UseCodeSite}
  FViewerType := wtFile;
  FDateTimeFormat := CodeSite.DateTimeFormat;
{$ELSE}
  FViewerType := wtOverseer;
  FDateTimeFormat := 'c';
{$ENDIF}
  FLogFilePrefix := 'ODB';
// Activate Debug Message
  SetEnabled(True);
// Simulate Attachments
  FSimulateAttachments := True;
end;

//
// ---------------------------------------------------------------------------------------
//

destructor TAsgDebugMessage.Destroy;
begin
  if (Self = DebugMessage) and not DebugMessageClosing then
  begin
    DbgSendError('Attempt to free the global DebugMessage!');
    Exit;
  end;
  CloseLogFile;
  FDateTimeFormat := '';
  FLogFilePrefix := '';
  inherited;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.OpenLogFile;
var
  FileName: string;
begin
  if FLogFileActive then Exit;
  FileName := FLogFilePrefix + '___';
  SetLength(FileName, 3);
  FileName := FileName + FormatDateTime('yymmdd', Date);
  Delete(FileName, 4, 1);
  FileName := FileName + '.log';
  AssignFile(FLogFile, FileName);
  try
    Rewrite(FLogFile);
    FLogFileActive := True;
  except
    FLogFileActive := False;
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.CloseLogFile;
begin
  if not FLogFileActive then Exit;
  Flush(FLogFile);
  try
    CloseFile(FLogFile);
  except                                                    // Eat exceptions
  end;
  FLogFileActive := False;
end;

//
// ---------------------------------------------------------------------------------------
//

function TAsgDebugMessage.GetEnabled: Boolean;
begin
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      Result := CodeSite.Enabled;
{$ENDIF}
    wtOverseer:
      Result := NxDebugging;
    wtGExperts:
// No way to determine status. An attemp to launch the viewer is done internally
// on every message send.
      Result := True;
// No way to determine status. Without viewer the message is eaten somewhere in
// the chain of Window's message handlers.
    wtDebugger:
      Result := True;
    wtConsole:
      Result := IsConsole;
    wtFile:
      Result := FLogFileActive;
  else
    Result := FWantEnabled;
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

function TAsgDebugMessage.CheckViewerType(Value: TAsgViewerType): Boolean;
begin
  case Value of
{$IFNDEF UseCodeSite}
    wtCodeSite:
      Result := False;
{$ENDIF}
    wtOverseer:
      Result := True;
    wtConsole:
      Result := IsConsole;
  else
    Result := True;
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

function TAsgDebugMessage.CheckEnabled: Boolean;
begin
  Result := FWantEnabled and (FViewerType <> wtIgnore) and Enabled
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.SetEnabled(Value: Boolean);
begin
  if (FWantEnabled <> Value) then
  begin
    FWantEnabled := Value;
    case FViewerType of
{$IFDEF UseCodeSite}
      wtCodeSite:
        CodeSite.Enabled := Value;
{$ENDIF}
      wtOverseer:
        if Value then
          NxStartDebug
        else
          NxEndDebug;
      wtFile:
        if Value then
          OpenLogFile
        else
          CloseLogFile;
    end;
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.SetDateTimeFormat(Value: string);
begin
  if AnsiCompareStr(FDateTimeFormat, Value) <> 0 then
  begin
    FDateTimeFormat := Value;
{$IFDEF UseCodeSite}
    CodeSite.DateTimeFormat := Value;
{$ENDIF}
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.SetViewerType(Value: TAsgViewerType);
var
  OldViewerType: TAsgViewerType;
  OldEnabled: Boolean;
begin
  if FViewerType <> Value then
  begin
    if not CheckViewerType(Value) then
      Exit;
    OldViewerType := FViewerType;
    OldEnabled := FWantEnabled;
    try
      Enabled := False;
      FViewerType := Value;
      InternalClear;
      case FViewerType of
        wtConsole:
          FSeparatorSize := 79;
        wtFile:
          FSeparatorSize := 120;
      else
        FSeparatorSize := 250;
      end;
    except
      FViewerType := OldViewerType;
      Enabled := OldEnabled;
    end;
  end;
end;

//
// =======================================================================================
//
// Basic Messages
//
// =======================================================================================
//
// Sends a dbgmInfo message to the Viewer. Use the DbgSendMsgEx method to specify a different
// message type.
//
// -----------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendMsg(const Msg: string);
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendMsg(Msg);
{$ENDIF}
    wtOverseer:
      Debugger.LogMsg(Msg);
  else
// Use extended version
    Self.DbgSendMsgEx(dbgmInfo, Msg);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

{$IFDEF UseCodeSite}

function TAsgDebugMessage.CodeSiteMsgType(AsgMsgType: Integer): Integer;
begin
  case AsgMsgType of
    dbgmInfo:
      Result := csmInfo;
    dbgmWarning:
      Result := csmWarning;
    dbgmError:
      Result := csmError;
    dbgmNote:
      Result := csmNote;
    dbgmCheckPoint:
      Result := csmCheckPoint;
    dbgmObject:
      Result := csmObject;
    dbgmStream:
      Result := csmStream;
    dbgmStringList:
      Result := csmStringList;
    dbgmProperty:
      Result := csmProperty;
    dbgmEnterMethod:
      Result := csmEnterMethod;
    dbgmExitMethod:
      Result := csmExitMethod;
    dbgmSeparator:
      Result := csmSeparator;
    dbgmInactiveSeparator:
      Result := csmInactiveSeparator;
    dbgmClear:
      Result := csmClear;
    dbgmScratchPad:
      Result := csmScratchPad;
    dbgmMemoryDump:
      Result := csmStringList;
  else
    Result := csmInfo;
  end;
end;

{$ENDIF}

//
// ---------------------------------------------------------------------------------------
//

function TAsgDebugMessage.OverseerMsgType(AsgMsgType: Integer): Integer;
begin
  case AsgMsgType of
    dbgmInfo:
      Result := dbgInfo;
    dbgmWarning:
      Result := dbgWarning;
    dbgmError:
      Result := dbgError;
    dbgmNote:
      Result := dbgNote;
    dbgmCheckPoint:
      Result := dbgCheckPoint;
    dbgmObject:
      Result := dbgObject;
    dbgmStream:
      Result := dbgStream;
    dbgmStringList:
      Result := dbgStringList;
    dbgmProperty:
      Result := dbgProperty;
    dbgmEnterMethod:
      Result := uDbgIntf.dbgEnterMethod;
    dbgmExitMethod:
      Result := uDbgIntf.dbgExitMethod;
    dbgmSeparator:
      Result := dbgSeparator;
    dbgmInactiveSeparator:
      Result := dbgInactiveSeparator;
    dbgmClear:
      Result := uDbgIntf.dbgClear;
    dbgmScratchPad:
      Result := dbgBlackBoard;
    dbgmMemoryDump:
      Result := dbgStringList;
  else
    Result := dbgInfo;
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

function TAsgDebugMessage.GExpertsMsgType(AsgMsgType: Integer): TMsgDlgType;
begin
  case AsgMsgType of
    dbgmInfo:
      Result := mtInformation;
    dbgmWarning:
      Result := mtWarning;
    dbgmError:
      Result := mtError;
    dbgmNote:
      Result := mtInformation;
  else
    Result := mtInformation;
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

function TAsgDebugMessage.ConsoleMsgType(AsgMsgType: Integer): string;
begin
  Result := StringOfChar('_', FIndentLevel);
  case AsgMsgType of
    dbgmInfo:
      Result := Result + 'Info: ';
    dbgmWarning:
      Result := Result + 'Warning: ';
    dbgmError:
      Result := Result + 'Error: ';
    dbgmNote:
      Result := Result + 'Note: ';
    dbgmCheckPoint:
      Result := Result + 'CheckPoint: ';
    dbgmObject:
      Result := Result + 'Object: ';
    dbgmStream:
      Result := Result + 'Stream: ';
    dbgmStringList:
      Result := Result + 'String List: ';
    dbgmProperty:
      Result := Result + 'Property: ';
    dbgmEnterMethod:
      Result := Result + 'Method --->> ';
    dbgmExitMethod:
      Result := Result + 'Method <<--- ';
    dbgmSeparator:
      Result := StringOfChar('-', FSeparatorSize);
    dbgmInactiveSeparator:
      Result := StringOfChar('#', FSeparatorSize);
    dbgmClear:
      Result := 'Clear Viewer';
    dbgmScratchPad:
      Result := 'Scratch Pad';
    dbgmmemoryDump:
      Result := 'Memory: ';
  else
    Result := Result + 'Info: ';
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendMsgEx(MsgType: Integer; const Msg: string);
var
  CombinedMsg: string;
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendMsgEx(CodeSiteMsgType(MsgType), Msg);
{$ENDIF}
    wtOverseer:
      Debugger.LogMsgEx(OverseerMsgType(MsgType), Msg);
    wtGExperts:
      begin
        if Length(Msg) = 0 then
          CombinedMsg := ConsoleMsgType(MsgType)
        else
          CombinedMsg := ConsoleMsgType(MsgType) + Msg;
        dbugIntf.SendDebugEx(CombinedMsg, GExpertsMsgType(MsgType))
      end;
    wtDebugger:
      begin
// ... Messages need to be terminated with a LineFeed to show up in the Debugger.
        CombinedMsg := ConsoleMsgType(MsgType) + Msg + #10;
        OutputDebugString(@CombinedMsg[1])
      end;
    wtConsole:
      begin
        CombinedMsg := ConsoleMsgType(MsgType) + Msg;
        WriteLn(CombinedMsg)
      end;
    wtFile:
      begin
// ... Fixed format to avoid possible exception in FormatDateTime
        CombinedMsg := FormatDateTime('c', now) + ' ' + ConsoleMsgType(MsgType) + Msg;
        WriteLn(FLogFile, CombinedMsg);
        Flush(FLogFile);
      end;
  else
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Sends a formatted dbgmInfo message to the CodeSite Viewer. The parameters to this
// message are the same as those used in the standard Format function. Use the DbgSendMsgEx
// method to specify a different message type.
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendFmtMsg(const Fmt: string; const Args: array of const);
var
  CombinedMsg: string;
begin
  if not CheckEnabled then Exit;
// ... the "Format" routine might throw exceptions
  try
    case FViewerType of
{$IFDEF UseCodeSite}
      wtCodeSite:
        CodeSite.SendFmtMsg(Fmt, Args);
{$ENDIF}
      wtOverseer:
        Debugger.LogFmtMsg(Fmt, Args);
    else
      begin
        CombinedMsg := Format(Fmt, Args);
        DbgSendMsg(CombinedMsg);
      end;
    end;
  except
    DbgSendError('Exception during "DbgSendFmtMsg"');
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendFmtMsgEx(MsgType: Integer; const Fmt: string; const Args: array of const);
var
  CombinedMsg: string;
begin
  if not CheckEnabled then Exit;
// ... the "Format" routine might throw exceptions
  try
    case FViewerType of
{$IFDEF UseCodeSite}
      wtCodeSite:
        CodeSite.SendFmtMsgEx(CodeSiteMsgType(MsgType), Fmt, Args);
{$ENDIF}
      wtOverseer:
        Debugger.LogFmtMsgEx(OverseerMsgType(MsgType), Fmt, Args);
    else
      begin
        CombinedMsg := Format(Fmt, Args);
        DbgSendMsgEx(MsgType, CombinedMsg);
      end;
    end;
  except
    DbgSendError('Exception during "DbgSendFmtMsgEx"');
  end;
end;

//
// =======================================================================================
//
// Derived Messages
//
// =======================================================================================
//
// Sends a dbgmNote message string to the Viewer. To format the message string, use the
// DbgSendFmtNote method which has the same parameters as the standard Format function.
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendNote(const Msg: string);
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendNote(Msg);
{$ENDIF}
    wtOverseer:
      Debugger.LogNote(Msg);
  else
// Use DbgSendMsgEx
    Self.DbgSendMsgEx(dbgmNote, Msg);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendFmtNote(const Fmt: string; const Args: array of const);
begin
  if not CheckEnabled then Exit;
// ... the "Format" routine might throw exceptions
  try
    case FViewerType of
{$IFDEF UseCodeSite}
      wtCodeSite:
        CodeSite.SendFmtNote(Fmt, Args);
{$ENDIF}
      wtOverseer:
        Debugger.LogFmtNote(Fmt, Args);
    else
// Use DbgSendFmtMsgEx
      Self.DbgSendFmtMsgEx(dbgmNote, Fmt, Args);
    end;
  except
    DbgSendError('Exception during "DbgSendFmtNote"');
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Sends a dbgmWarning message string to the Viewer. To format the message string, use
// the DbgSendFmtNote method which has the same parameters as the standard Format function.
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendWarning(const Msg: string);
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendWarning(Msg);
{$ENDIF}
    wtOverseer:
      Debugger.LogWarning(Msg);
  else
// Use DbgSendMsgEx
    Self.DbgSendMsgEx(dbgmWarning, Msg);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendFmtWarning(const Fmt: string; const Args: array of const);
begin
  if not CheckEnabled then Exit;
// ... the "Format" routine might throw exceptions
  try
    case FViewerType of
{$IFDEF UseCodeSite}
      wtCodeSite:
        CodeSite.SendFmtWarning(Fmt, Args);
{$ENDIF}
      wtOverseer:
        Debugger.LogFmtWarning(Fmt, Args);
    else
// Use DbgSendFmtMsgEx
      Self.DbgSendFmtMsgEx(dbgmWarning, Fmt, Args);
    end;
  except
    DbgSendError('Exception during "DbgSendFmtWarning"');
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Sends a dbgmError message string to the Viewer. To format the message string, use the
// DbgSendFmtError method which has the same parameters as the standard Format function.
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendError(const Msg: string);
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendError(Msg);
{$ENDIF}
    wtOverseer:
      Debugger.LogError(Msg);
  else
// Use DbgSendMsgEx
    Self.DbgSendMsgEx(dbgmError, Msg);
  end;
end;
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendFmtError(const Fmt: string; const Args: array of const);
begin
  if not CheckEnabled then Exit;
// ... the "Format" routine might throw exceptions
  try
    case FViewerType of
{$IFDEF UseCodeSite}
      wtCodeSite:
        CodeSite.SendFmtError(Fmt, Args);
{$ENDIF}
      wtOverseer:
        Debugger.LogFmtError(Fmt, Args);
    else
// Use DbgSendFmtMsgEx
      Self.DbgSendFmtMsgEx(dbgmError, Fmt, Args);
    end;
  except
    DbgSendError('Exception during "DbgSendFmtError"');
  end;
end;

//
// =======================================================================================
//
// Viewer Control Messages
//
// =======================================================================================
//
// Sends a dbgmCheckPoint message to the Viewer. The check point index is automatically
// incremented when the message is sent. Check points are useful when you need to keep
// track of how many times a particular section of code is being executed. To add just
// a separator bar to the message list, use the AddSeparator method.
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgAddCheckPoint(const CheckName: string);
begin
  if not CheckEnabled then Exit;
  Inc(FCheckPointCount);
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.AddCheckPoint;
{$ENDIF}
    wtOverseer:
      Debugger.AddCheckPoint(Format('-%d- %s', [FCheckPointCount, CheckName]));
  else
// Use DbgSendFmtMsgEx
    Self.DbgSendMsgEx(dbgmCheckPoint, Format('-%d- %s', [FCheckPointCount, CheckName]));
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Sends a dbgmSeparator message to the Viewer. Separators are displayed as thin gray
// lines in the message list and are used to group messages. To associate an index with
// a separator, use the AddCheckPoint method.
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgAddSeparator;
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.AddSeparator;
{$ENDIF}
    wtOverseer:
      Debugger.AddSeparator;
  else
// Use DbgSendMsgEx
    Self.DbgSendMsgEx(dbgmSeparator, '');
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Instructs the Viewer to clear the message list. When the Clear method is called, the
// user is not asked to confirm the operation even if the Display clear confirmation
// dialog option is checked. Note that when the user elects to clear the message list
// from within the Viewer, the confirmation dialog option is honored.
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.InternalClear;
begin
  FCheckPointCount := 0;
  FIndentLevel := 0;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgClear;
begin
  if not CheckEnabled then Exit;
  InternalClear;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.Clear;
{$ENDIF}
    wtOverseer:
      Debugger.Clear;
    wtGExperts:
      dbugIntf.ClearDebug;
    wtFile:
      begin
        CloseLogFile;
        OpenLogFile;
      end
  else
// Use DbgSendFmtMsgEx
    Self.DbgSendMsgEx(dbgmClear, '');
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Sends a dbgmEnterMethod message to the Viewer. This instructs the viewer to increment
// the current indent level, which affects all future messages until a dbgExitMethod
// message is received or the message list is cleared. Call the DbgEnterMethod method at the
// begin of a method, procedure, or function to enable Call Stack tracking in the Viewer.
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgEnterMethod(const Msg: string);
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.EnterMethod(Msg);
{$ENDIF}
    wtOverseer:
      Debugger.EnterProc(Msg);
  else
// Use DbgSendMsgEx
    Self.DbgSendMsgEx(dbgmEnterMethod, Msg);
  end;
  FIndentLevel := FIndentLevel + 2;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgEnterMethodEx(const Msg: string; const Value: string);
var
  CombinedMsg: string;
begin
  CombinedMsg := Msg + ': ' + Value;
  DbgEnterMethod(CombinedMsg);
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgEnterMethodFmtEx(const Msg: string; const Fmt: string; const Args: array of const);
var
  CombinedMsg: string;
begin
// ... The Format routine might throw exceptions
  try
    CombinedMsg := Msg + ': ' + Format(Fmt, Args);
    DbgEnterMethod(CombinedMsg);
  except
    DbgSendError('Exception during "DbgEnterMethodFmtEx"');
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Sends a dbgmExitMethod message to the Viewer. This instructs the viewer to decrement
// the current indent level, which affects all future messages until a dbgmEnterMethod
// message is received or the message list is cleared. Call the DbgExitMethod method at
// the end of a method, procedure, or function to enable Call Stack tracking in the Viewer.
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgExitMethod(const Msg: string);
begin
  if not CheckEnabled then Exit;
  FIndentLevel := FIndentLevel - 2;
  if FIndentLevel < 0 then
    FIndentLevel := 0;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.ExitMethod(Msg);
{$ENDIF}
    wtOverseer:
      Debugger.LeaveProc(Msg);
  else
// Use DbgSendMsgEx
    Self.DbgSendMsgEx(dbgmExitMethod, Msg);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgExitMethodEx(const Msg: string; const Value: string);
var
  CombinedMsg: string;
begin
  CombinedMsg := Msg + ': ' + Value;
  DbgExitMethod(CombinedMsg);
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgExitMethodFmtEx(const Msg: string; const Fmt: string; const Args: array of const);
var
  CombinedMsg: string;
begin
// ... The Format routine might throw exceptions
  try
    CombinedMsg := Msg + ': ' + Format(Fmt, Args);
    DbgExitMethod(CombinedMsg);
  except
    DbgSendError('Exception during "DbgExitMethodFmtEx"');
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Sends non-persistent string message to Scratch Pad pane of the Viewer. The PadIndex
// parameter is used to specify which scratch pad will receive and display the message.
// There are 100 scratch pad lines available (0..99). To format the message string, use
// the DbgFmtScratchPad method, which has the same parameters as the standard Format function.
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgScratchPad(LineIndex: Integer; const Msg: string);
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.ScratchPad(LineIndex, Msg);
{$ENDIF}
    wtOverseer:
      Debugger.ScratchPad(LineIndex, Msg);
  else
// Use DbgSendMsgEx
//    Self.DbgSendMsgEx(dbgmScratchPad, Msg);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgFmtScratchPad(LineIndex: Integer; const Fmt: string; const Args: array of const);
begin
  if not CheckEnabled then Exit;
// ... The Format routine might throw exceptions
  try
    case FViewerType of
{$IFDEF UseCodeSite}
      wtCodeSite:
        CodeSite.FmtScratchPad(LineIndex, Fmt, Args);
{$ENDIF}
      wtOverseer:
        Debugger.FmtScratchPad(LineIndex, Fmt, Args);
    else
// Use DbgSendFmtMsgEx
//      Self.DbgSendFmtMsgEx(dbgmScratchPad, Fmt, Args);
    end;
  except
    DbgSendError('Exception during "DbgEnterMethodFmtEx"');
  end;
end;

//
// =======================================================================================
//
// Computed Messages
//
// =======================================================================================
//
// Tests whether the Value parameter is assigned and then passes the result as a dbgmInfo
// message to the Viewer. Use the DbgSendAssignedEx method to specify a different message
// type.
//
// .......................................................................................
//
//  CodeSite.SendAssigned ( 'B', B );
//
//  B is not assigned.
//  B is assigned.
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendAssigned(const Msg: string; Value: Pointer);
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendAssigned(Msg, Value);
{$ENDIF}
    wtOverseer:
      Debugger.LogAssigned(Msg, Value);
  else
// Use the extended version
    Self.DbgSendAssignedEx(dbgmInfo, Msg, Value);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendAssignedEx(MsgType: Integer; const Msg: string; Value: Pointer);
var
  ComputedMsg: string;
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendAssignedEx(CodeSiteMsgType(MsgType), Msg, Value);
{$ENDIF}
    wtOverseer:
      Debugger.LogAssignedEx(OverseerMsgType(MsgType), Msg, Value);
  else
// Simulate the CodeSite method
    begin
      if Assigned(Value) then
        ComputedMsg := Msg + ' is assigned.'
      else
        ComputedMsg := Msg + ' is not assigned.';
      Self.DbgSendMsgEx(MsgType, ComputedMsg);
    end;
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Evaluates the Expression parameter and then passes the result as a dbgmInfo message to
// the CodeSite Viewer. Use the DbgSendBooleanEx method to specify a different message
// type.
//
// .......................................................................................
//
// CodeSite.SendBoolean ( 'Edit1.Enabled:', Edit1.Enabled );
// CodeSite.SendBoolean ( 'X is 12', X = 12 );
//
//  Edit1.Enabled: True
//  X is 12 = False
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendBoolean(const Msg: string; Expression: Boolean);
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendBoolean(Msg, Expression);
{$ENDIF}
    wtOverseer:
      Debugger.LogBoolean(Msg, Expression);
  else
// Use the extended version
    Self.DbgSendBooleanEx(dbgmInfo, Msg, Expression);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendBooleanEx(MsgType: Integer; const Msg: string; Expression: Boolean);
var
  ComputedMsg: string;
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendBooleanEx(CodeSiteMsgType(MsgType), Msg, Expression);
{$ENDIF}
    wtOverseer:
      Debugger.LogBooleanEx(OverseerMsgType(MsgType), Msg, Expression);
  else
// Simulate the CodeSite method
    begin
      if Msg[Length(Msg)] = ':' then
        ComputedMsg := Msg + ' '
      else
        ComputedMsg := Msg + ' = ';
      if Expression then
        ComputedMsg := ComputedMsg + 'True'
      else
        ComputedMsg := ComputedMsg + 'False';
      Self.DbgSendMsgEx(MsgType, ComputedMsg);
    end;
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Determines the color symbol name (if available) of the Value parameter and the RGB
// values for the corresponding color and sends the formatted result as a dbgmInfo message
// to the Viewer. Use the DbgSendColorEx method to specify a different message type.
//
// .......................................................................................
//
// CodeSite.SendColor ( 'Panel1.Color', Panel1.Color );
// CodeSite.SendColor ( 'Label1.Font.Color', Label1.Font.Color );
//
// Panel1.Color: clBtnFace(192, 192, 192);
// Label1.Font.Color: $00660033(51, 0, 102);
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendColor(const Msg: string; Value: TColor);
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendColor(Msg, Value);
{$ENDIF}
    wtOverseer:
      Debugger.LogColor(Msg, Value);
  else
// Use the extended version
    Self.DbgSendColorEx(dbgmInfo, Msg, Value);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendColorEx(MsgType: Integer; const Msg: string; Value: TColor);
var
  ComputedMsg: string;
  N: Integer;
  B: Integer;
  R: Integer;
  G: Integer;
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendColorEx(CodeSiteMsgType(MsgType), Msg, Value);
{$ENDIF}
    wtOverseer:
      Debugger.LogColorEx(OverseerMsgType(MsgType), Msg, Value);
  else
// Simulate the CodeSite method
    begin
      N := ColorToRGB(Value);
      R := N div $10000;
      G := (N mod $10000) div $100;
      B := N mod $100;
      ComputedMsg := Msg + ': ' + ColorToString(Value) + '(' + IntToStr(B) + ', '
        + IntToStr(G) + ', ' + IntToStr(R) + ')';
      Self.DbgSendMsgEx(MsgType, ComputedMsg);
    end;
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Formats the Value parameter as a string and then passes the result as a dbgmInfo message
// to the Viewer. Use the DbgSendIntegerEx method to specify a different message type. To
// send floating-point numbers to the Viewer, use the DbgSendFloat method.
//
// .......................................................................................
//
// CodeSite.SendInteger ( 'Width', Width )
//
// Width = 200
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendInteger(const Msg: string; Value: Longint);
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendInteger(Msg, Value);
{$ENDIF}
    wtOverseer:
      Debugger.LogInteger(Msg, Value);
  else
// Use the extended version
    Self.DbgSendIntegerEx(dbgmInfo, Msg, Value);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendIntegerEx(MsgType: Integer; const Msg: string; Value: Longint);
var
  ComputedMsg: string;
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendIntegerEx(CodeSiteMsgType(MsgType), Msg, Value);
{$ENDIF}
    wtOverseer:
      Debugger.LogIntegerEx(OverseerMsgType(MsgType), Msg, Value);
  else
// Simulate the CodeSite method
    begin
      ComputedMsg := Msg + ' = ' + IntToStr(Value);
      Self.DbgSendMsgEx(MsgType, ComputedMsg);
    end;
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Formats the Value parameter as a string and then passes the result as a dbgmInfo
// message to the Viewer. Use the DbgSendPointEx method to specify a different message
// type.
//
// .......................................................................................
//
// CodeSite.SendPoint ( 'BoundsRect.TopLeft', BoundsRect.TopLeft );
//
// BoundsRect.TopLeft = (200, 108)
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendPoint(const Msg: string; Value: TPoint);
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendPoint(Msg, Value);
{$ENDIF}
    wtOverseer:
      Debugger.LogPoint(Msg, Value);
  else
// Use the extended version
    Self.DbgSendPointEx(dbgmInfo, Msg, Value);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendPointEx(MsgType: Integer; const Msg: string; Value: TPoint);
var
  ComputedMsg: string;
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendPointEx(CodeSiteMsgType(MsgType), Msg, Value);
{$ENDIF}
    wtOverseer:
      Debugger.LogPointEx(OverseerMsgType(MsgType), Msg, Value);
  else
// Simulate the CodeSite method
    begin
      ComputedMsg := Msg + ' = (' + IntToStr(Value.X) + ', ' + IntToStr(Value.Y) + ')';
      Self.DbgSendMsgEx(MsgType, ComputedMsg);
    end;
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Formats the Value parameter as a string and then passes the result as a dbgmInfo
// message to the Viewer. Use the DbgSendRectEx method to specify a different message type.
//
// .......................................................................................
//
// CodeSite.SendRect ( 'BoundsRect', BoundsRect );
//
// BoundsRect = (200, 108, 460, 320)
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendRect(const Msg: string; Value: TRect);
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendRect(Msg, Value);
{$ENDIF}
    wtOverseer:
      Debugger.LogRect(Msg, Value);
  else
// Use the extended version
    Self.DbgSendRectEx(dbgmInfo, Msg, Value);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendRectEx(MsgType: Integer; const Msg: string; Value: TRect);
var
  ComputedMsg: string;
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendRectEx(CodeSiteMsgType(MsgType), Msg, Value);
{$ENDIF}
    wtOverseer:
      Debugger.LogRectEx(OverseerMsgType(MsgType), Msg, Value);
  else
// Simulate the CodeSite method
    begin
      ComputedMsg := Msg + ' = (' + IntToStr(Value.Left) + ', ' + IntToStr(Value.Top)
        + ', ' + IntToStr(Value.Right) + ', ' + IntToStr(Value.Bottom) + ')';
      Self.DbgSendMsgEx(MsgType, ComputedMsg);
    end;
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Formats the Value parameter as a string and then passes the result as a dbgmInfo message
// to the Viewer. Use the DbgSendStringEx method to specify a different message type.
//
// .......................................................................................
//
// CodeSite.SendString ( 'Caption', Caption );
//
// Caption = 'Form1'
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendString(const Msg: string; const Value: string);
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendString(Msg, Value);
{$ENDIF}
    wtOverseer:
      Debugger.LogString(Msg, Value);
  else
// Use the extended version
    Self.DbgSendStringEx(dbgmInfo, Msg, Value);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendStringEx(MsgType: Integer; const Msg: string; const Value: string);
var
  ComputedMsg: string;
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendStringEx(CodeSiteMsgType(MsgType), Msg, Value);
{$ENDIF}
    wtOverseer:
      Debugger.LogStringEx(OverseerMsgType(MsgType), Msg, Value);
  else
// Simulate the CodeSite method
    begin
      ComputedMsg := Msg + ' = ' + '''' + Value + '''';
      Self.DbgSendMsgEx(MsgType, ComputedMsg);
    end;
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Formats the Value parameter as a string and then passes the result as a dbgmInfo message
// to the Viewer. Use the DbgSendCurrencyEx method to specify a different message type.
// The format of the currency value is controlled by the Regional Settings in the
// Control Panel.
//
// .......................................................................................
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendCurrency(const Msg: string; Value: Currency);
begin
  if not CheckEnabled then Exit;
{$IFDEF UseCodeSite}
  if FViewerType = wtCodeSite then
    CodeSite.SendCurrency(Msg, Value)
  else
{$ENDIF}
// Use the extended version
    Self.DbgSendCurrencyEx(dbgmInfo, Msg, Value);
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendCurrencyEx(MsgType: Integer; const Msg: string; Value: Currency);
var
  ComputedMsg: string;
begin
  if not CheckEnabled then Exit;
{$IFDEF UseCodeSite}
  if FViewerType = wtCodeSite then
    CodeSite.SendCurrencyEx(CodeSiteMsgType(MsgType), Msg, Value)
  else
{$ENDIF}
// Simulate the CodeSite method
  begin
    ComputedMsg := Msg;
    ComputedMsg := Msg + ' = ' + FormatCurr('n', Value);
    Self.DbgSendMsgEx(MsgType, ComputedMsg);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Formats the Value parameter as a string and then passes the result as a dbgmInfo message
// to the Viewer. Use the DbgSendFloatEx method to specify a different message type. The
// value is formatted according to the rules of the "n" format string.
//
// .......................................................................................
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendFloat(const Msg: string; Value: Extended);
begin
  if not CheckEnabled then Exit;
{$IFDEF UseCodeSite}
  if FViewerType = wtCodeSite then
    CodeSite.SendFloat(Msg, Value)
  else
{$ENDIF}
// Use the extended version
    Self.DbgSendFloatEx(dbgmInfo, Msg, Value);
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendFloatEx(MsgType: Integer; const Msg: string; Value: Extended);
var
  ComputedMsg: string;
begin
  if not CheckEnabled then Exit;
{$IFDEF UseCodeSite}
  if FViewerType = wtCodeSite then
    CodeSite.SendFloatEx(CodeSiteMsgType(MsgType), Msg, Value)
  else
{$ENDIF}
// Simulate the CodeSite method
  begin
    ComputedMsg := Msg + ' = ' + FormatFloat('n', Value);
    Self.DbgSendMsgEx(MsgType, ComputedMsg);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Formats the Value parameter as a string and then passes the result as a dbgmInfo
// message to the Viewer. Use the DbgSendDateTimeEx method to specify a different message
// type. The DateTimeFormat property controls the format used to build the string
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendDateTime(const Msg: string; Value: TDateTime);
begin
  if not CheckEnabled then Exit;
{$IFDEF UseCodeSite}
  if FViewerType = wtCodeSite then
    CodeSite.SendDateTime(Msg, Value)
  else
{$ENDIF}
// Use the extended version
    Self.DbgSendDateTimeEx(dbgmInfo, Msg, Value);
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendDateTimeEx(MsgType: Integer; const Msg: string; Value: TDateTime);
var
  ComputedMsg: string;
begin
  if not CheckEnabled then Exit;
{$IFDEF UseCodeSite}
  if FViewerType = wtCodeSite then
    CodeSite.SendDateTimeEx(CodeSiteMsgType(MsgType), Msg, Value)
  else
{$ENDIF}
// Simulate the CodeSite method
  begin
    ComputedMsg := Msg + ' = ' + FormatDateTime(FDateTimeFormat, Value);
    Self.DbgSendMsgEx(MsgType, ComputedMsg);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Formats the Value parameter as a string and then passes the result as a dbgmInfo message
// to the Viewer. Use the DbgSendDWordEx method to specify a different message type. Under
// Delphi 3, DWord maps to a 32-bit signed Integer. Under Delphi 4, DWord maps to a
// LongWord--a 32-bit unsigned Integer.
//
// .......................................................................................
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendDWord(const Msg: string; Value: DWord);
begin
  if not CheckEnabled then Exit;
{$IFDEF UseCodeSite}
  if FViewerType = wtCodeSite then
    CodeSite.SendDWord(Msg, Value)
  else
{$ENDIF}
// Use the extended version
    Self.DbgSendDWordEx(dbgmInfo, Msg, Value);
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendDWordEx(MsgType: Integer; const Msg: string; Value: DWord);
var
  ComputedMsg: string;
begin
  if not CheckEnabled then Exit;
{$IFDEF UseCodeSite}
  if FViewerType = wtCodeSite then
    CodeSite.SendDWordEx(CodeSiteMsgType(MsgType), Msg, Value)
  else
{$ENDIF}
// Simulate the CodeSite method
  begin
    ComputedMsg := Msg + ' = ' + IntToStr(Value);
    Self.DbgSendMsgEx(MsgType, ComputedMsg);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Formats the Value parameter as a string and then passes the result as a dbgmInfo message
// to the Viewer. Use the DbgSendNumAsHex method to specify a different message type. Under
// Delphi 3, DWord maps to a 32-bit signed Integer. Under Delphi 4, DWord maps to a
// LongWord--a 32-bit unsigned Integer. Instead of being formatted as a normal base-10
// number, the value is formatted as a hexidecimal number. Under Delphi, the hex value is
// prefixed with a "$", while under C++Builder the hex value is prefixed with a "0x".
//
// .......................................................................................
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendNumAsHex(const Msg: string; Value: DWord);
begin
  if not CheckEnabled then Exit;
{$IFDEF UseCodeSite}
  if FViewerType = wtCodeSite then
    CodeSite.SendNumAsHex(Msg, Value)
  else
{$ENDIF}
// Use the extended version
    Self.DbgSendNumAsHexEx(dbgmInfo, Msg, Value);
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendNumAsHexEx(MsgType: Integer; const Msg: string; Value: DWord);
var
  ComputedMsg: string;
begin
  if not CheckEnabled then Exit;
{$IFDEF UseCodeSite}
  if FViewerType = wtCodeSite then
    CodeSite.SendNumAsHexEx(CodeSiteMsgType(MsgType), Msg, Value)
  else
{$ENDIF}
// Simulate the CodeSite method
  begin
    ComputedMsg := Msg + ' = $' + IntToHex(Value, 0);
    Self.DbgSendMsgEx(MsgType, ComputedMsg);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Formats the Value parameter as a string and then passes the result as a dbgmInfo
// message to the Viewer. Use the DbgSendEnumEx method to specify a different message type.
// The TypeInfo function is used to pass the RTTI for the desired enumeration.
//
// .......................................................................................
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendEnum(const Msg: string; EnumTypeInfo: PTypeInfo; Value: Integer);
begin
  if not CheckEnabled then Exit;
{$IFDEF UseCodeSite}
  if FViewerType = wtCodeSite then
    CodeSite.SendEnum(Msg, EnumTypeInfo, Value)
  else
{$ENDIF}
// Use the extended version
    Self.DbgSendEnumEx(dbgmInfo, Msg, EnumTypeInfo, Value);
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendEnumEx(MsgType: Integer; const Msg: string; EnumTypeInfo: PTypeInfo; Value: Integer);
var
  ComputedMsg: string;
begin
  if not CheckEnabled then Exit;
{$IFDEF UseCodeSite}
  if FViewerType = wtCodeSite then
    CodeSite.SendEnumEx(CodeSiteMsgType(MsgType), Msg, EnumTypeInfo, Value)
  else
{$ENDIF}
// Simulate the CodeSite method
  begin
    ComputedMsg := Msg + ' = ' + GetEnumName(EnumTypeInfo, Value);
    Self.DbgSendMsgEx(MsgType, ComputedMsg);
  end;
end;

//
// =======================================================================================
//
// Messages with Attachments
//
// =======================================================================================
//
// Sends a dbgmObject message to the CodeSite Viewer. The Msg string is displayed in the
// message list, while a description of the published properties and their current values
// for the Obj parameter is displayed in the inspector pane.
//
// If the viewer dosn't support attachments, the attachment can be promoted to
// additional messages by setting the SimulateAttachments property.
//
// ---------------------------------------------------------------------------------------
//

function TAsgDebugMessage.ObjectToStringList(Obj: TObject): TStringList;

  procedure AddObjectToStringList(PropOwner: TObject; List: TStringList; Level: Integer);
  type
    TIntegerSet = set of 0..SizeOf(Integer) * 8 - 1;        // see CLASSES.PAS
  var
    PropIdx: integer;
    PropertyList: ^TPropList;
    PropertyName: string;
    PropertyInfo: PPropInfo;
    PropertyType: PTypeInfo;
    PropertyKind: TTypeKind;
    BaseType: PTypeInfo;
    BaseData: PTypeData;
    GetProc: Pointer;
    OrdValue: Integer;
    FloatValue: Extended;
    N: Integer;
//  B: Integer;
//  R: Integer;
//  G: Integer;
    Prefix: string;
    NewLine: string;
    EnumName: string;
    NextObject: TObject;
    FollowObject: Boolean;
  begin
    GetMem ( PropertyList, SizeOf(TPropList) );
    try
      Prefix := StringOfChar(' ', 2 * Level);
// Build list of published properties
      FillChar(PropertyList^[0], SizeOf(TPropList), #00);
      GetPropList(PropOwner.ClassInfo, tkProperties - [tkArray, tkRecord, tkInterface], @PropertyList^[0]);
// Process property list
      PropIdx := 0;
      while ((PropIdx < High(PropertyList^)) and (nil <> PropertyList^[PropIdx])) do
      begin
// Get information about found properties
        PropertyInfo := PropertyList^[PropIdx];
        PropertyType := PropertyInfo^.PropType^;
        PropertyKind := PropertyType^.Kind;
        PropertyName := PropertyInfo^.Name;
// Write only property
        GetProc := PropertyInfo^.GetProc;
        if not Assigned(GetProc) then
        begin
          NewLine := Prefix + '  ' + PropertyName + ' = <' + PropertyType^.Name + '> (can''t be read)';
          List.Add(NewLine);
        end
        else
        begin
          case PropertyKind of
// Set
            tkSet:
              begin
                BaseType := GetTypeData(PropertyType)^.CompType^;
                BaseData := GetTypeData(BaseType);
                OrdValue := GetOrdProp(PropOwner, PropertyInfo);
                NewLine := Prefix + '+ ' + PropertyName + ' = [' + BaseType^.Name + ']';
                List.Add(NewLine);
                for N := BaseData^.MinValue to BaseData^.MaxValue do
                begin
                  EnumName := GetEnumName(BaseType, N);
                  if EnumName = '' then
                    Break;
                  NewLine := Prefix + '    ' + EnumName;
                  if N in TIntegerSet(OrdValue) then
                    NewLine := NewLine + ' = True'
                  else
                    NewLine := NewLine + ' = False';
                  List.Add(NewLine);
                end;
              end;
// Integer
            tkInteger:
              begin
                OrdValue := GetOrdProp(PropOwner, PropertyInfo);
// Special case: colors
//              if ColorToIdent(OrdValue, StringValue) then
//              begin
//                N := ColorToRGB(OrdValue);
//                R := N div $10000;
//                G := (N mod $10000) div $100;
//                B := N mod $100;
//                NewLine := Prefix + '  ' + PropertyName + ' = ' + StringValue + '(' + IntToStr(B) + ', ' + IntToStr(G) + ', ' + IntToStr(R) + ')';
//              end
// Normal integer      ??? 0 = clBlack ???
//              else
                NewLine := Prefix + '  ' + PropertyName + ' = ' + IntToStr(OrdValue);
                List.Add(NewLine);
              end;
// Single 8 bit character (ANSI)
            tkChar:
              begin
                OrdValue := GetOrdProp(PropOwner, PropertyInfo);
                NewLine := Prefix + '  ' + PropertyName + ' = ' + '#$' + IntToHex(OrdValue, 2);
                List.Add(NewLine);
              end;
// Single 16 bit character (UNICODE)
            tkWChar:
              begin
                OrdValue := GetOrdProp(PropOwner, PropertyInfo);
                NewLine := Prefix + '  ' + PropertyName + ' = #$' + IntToHex(OrdValue, 4);
                List.Add(NewLine);
              end;
// Class
            tkClass:
              begin
                OrdValue := GetOrdProp(PropOwner, PropertyInfo);
                if OrdValue = 0 then
                begin
                  NewLine := Prefix + '  ' + PropertyName + ' = <' + PropertyType^.Name + '> (not assigned)';
                  List.Add(NewLine);
                end
                else
                begin
                  NextObject := TObject(OrdValue);
                  NewLine := Prefix + '  ' + PropertyName + ' = <' + PropertyType^.Name + '>';
                  if NextObject is TComponent then
                  begin
                     FollowObject := False;
                     NewLine := NewLine + ': ' + TComponent(NextObject).Name
                  end
                  else
                  begin
                    FollowObject := True;
                    NewLine[Succ(Length(Prefix))] := '*';
                  end;
                  List.Add(NewLine);
                  if FollowObject then
                  begin
                    try
                      AddObjectToStringList(NextObject, List, Level + 1);
                    except
                      List.Add('*** Exception triggered ***' );
                      end;
                  end;
                end;
              end;
// Floating point value
            tkFloat:
              begin
                FloatValue := GetFloatProp(PropOwner, PropertyInfo);
                NewLine := Prefix + '  ' + PropertyName + ' = ' + FormatFloat('n', FloatValue);
                List.Add(NewLine);
              end;
// Enumerated value
            tkEnumeration:
              begin
                OrdValue := GetOrdProp(PropOwner, PropertyInfo);
                NewLine := Prefix + '  ' + PropertyName + ' = ' + GetEnumName(PropertyType, OrdValue);
                List.Add(NewLine);
              end;
// Long Ansi string
            tkLString:
              begin
                NewLine := Prefix + '  ' + PropertyName + ' = ' + '''' + GetStrProp(PropOwner, PropertyInfo) + '''';
                List.Add(NewLine);
              end;
// Long UNICODE string
            tkWString:
              begin
                NewLine := Prefix + '  ' + PropertyName + ' = ' + '''' + GetStrProp(PropOwner, PropertyInfo) + '''';
                List.Add(NewLine);
              end;
// Short string
            tkString:
              begin
                NewLine := Prefix + '  ' + PropertyName + ' = ' + '''' + GetStrProp(PropOwner, PropertyInfo) + '''';
                List.Add(NewLine);
              end;
// Variant
            tkVariant:
              begin
                NewLine := Prefix + '  ' + PropertyName + ' = ' + GetVariantProp(PropOwner, PropertyInfo);
                List.Add(NewLine);
              end;
// Method
            tkMethod:
              begin
                NewLine := Prefix + '  ' + PropertyName + ' = (' + GetEnumName(TypeInfo(TMethodKind), Ord(GetTypeData(PropertyType)^.MethodKind)) + ')';
                List.Add(NewLine);
              end;
// All others
          else
            begin
              NewLine := Prefix + '  ' + PropertyName + ' = <' + PropertyType^.Name + '> ('
                + GetEnumName(TypeInfo(TTypeKind), Ord(PropertyKind)) + ')';
              List.Add(NewLine);
            end;
          end
        end;
// Next item in the property list
        Inc(PropIdx);
      end;
      NewLine := '';
    finally
      if NewLine <> '' then
        List.Add(NewLine);
      FreeMem(PropertyList);
      end;
  end;

begin
  Result := TStringList.Create;
  try
    AddObjectToStringList(Obj, Result, 0)
  except
    Result.Add('*** Exception triggered ***' );
    end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendObject(const Msg: string; Obj: TObject);
var
  List: TStringList;
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendObject(Msg, Obj);
{$ENDIF}
// Functionality is missing
//  wtOverseer:
//    Debugger.LogObject(Msg, Obj);
// Simulate the CodeSite method
  else
    if not Assigned(obj) then
      Self.DbgSendError(Msg + ' (Object is not assigned)')
    else
    begin
      List := ObjectToStringList(Obj);
      try
        if List.Count = 0 then
          Self.DbgSendMsgEx(dbgmObject, Msg + ' (Object has no published properties)')
        else if SimulateAttachments then
          Self.DbgSendStringListEx(dbgmObject, Msg, List)
        else
          Self.DbgSendMsgEx(dbgmObject, Msg + ' (Attachments not supported)');
      finally
        List.Free;
      end;
    end
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Uses RTTI to determine if a property named PropName exists for Obj. If one does, then
// the type of the property is determined so that the current value of the property can
// be correctly displayed. If the property is a class type, then a call to the
// DbgSendObject method. Otherwise, the current value of the property is formatted along
// with the Msg string and is sent to the Viewer as a dbgmProperty message.
//
// If the viewer dosn't support attachments, the attachment can be promoted to
// additional messages by setting the SimulateAttachments property.
//
// ---------------------------------------------------------------------------------------
//

function TAsgDebugMessage.PropertyToString(PropOwner: TObject; PropName: string; var PropObject: TObject): string;
type
  TIntegerSet = set of 0..SizeOf(Integer) * 8 - 1;          // see CLASSES.PAS
var
  PropertyInfo: PPropInfo;
  PropertyType: PTypeInfo;
  PropertyKind: TTypeKind;
  BaseType: PTypeInfo;
  BaseData: PTypeData;
  GetProc: Pointer;
  OrdValue: Integer;
  FloatValue: Extended;
  StringValue: string;
  EnumName: string;
  N: Integer;
//B: Integer;
  R: Integer;
//G: Integer;
begin
  PropObject := nil;
  PropertyInfo := GetPropInfo(PropOwner.ClassInfo, PropName);
  if Assigned(PropertyInfo) then
  begin
    PropertyType := PropertyInfo^.PropType^;
    GetProc := PropertyInfo^.GetProc;
    PropertyKind := PropertyType^.Kind;
    Result := '<' + PropertyType^.Name + '>';
    if not Assigned(GetProc) then
      Result := Result + ' (can''t be read)'
    else
    begin
      case PropertyKind of
        tkSet:
          begin
            BaseType := GetTypeData(PropertyType)^.CompType^;
            BaseData := GetTypeData(BaseType);
            OrdValue := GetOrdProp(PropOwner, PropertyInfo);
            Result := '[';
            R := 0;
            for N := BaseData^.MinValue to BaseData^.MaxValue do
              if N in TIntegerSet(OrdValue) then
              begin
                EnumName := GetEnumName(BaseType, N);
                if EnumName = '' then
                  Break;
                if R > 0 then
                  Result := Result + ', ';
                Result := Result + EnumName;
                inc(R);
              end;
            Result := Result + ']';
          end;
        tkInteger:
          begin
            OrdValue := GetOrdProp(PropOwner, PropertyInfo);
//          if ColorToIdent(OrdValue, StringValue) then
//          begin
//            N := ColorToRGB(OrdValue);
//            R := N div $10000;
//            G := (N mod $10000) div $100;
//            B := N mod $100;
//            Result := StringValue + '(' + IntToStr(B) + ', ' + IntToStr(G) + ', ' + IntToStr(R) + ')';
//          end
//          else       ??? 0 --> clBlack ???
              Result := IntToStr(OrdValue);
          end;
        tkChar:
          begin
            OrdValue := GetOrdProp(PropOwner, PropertyInfo);
            Result := '#$' + IntToHex(OrdValue, 2);
          end;
        tkWChar:
          begin
            OrdValue := GetOrdProp(PropOwner, PropertyInfo);
            Result := '#$' + IntToHex(OrdValue, 4);
          end;
        tkClass:
          begin
            OrdValue := GetOrdProp(PropOwner, PropertyInfo);
            Result := '<' + PropertyType^.Name + '>';
            if OrdValue = 0 then
              Result := Result + ' (not assigned)'
            else
              Result := Result + ' (assigned)';
            PropObject := TObject(OrdValue);
          end;
        tkFloat:
          begin
            FloatValue := GetFloatProp(PropOwner, PropertyInfo);
            Result := FormatFloat('n', FloatValue);
          end;
        tkEnumeration:
          begin
            OrdValue := GetOrdProp(PropOwner, PropertyInfo);
            Result := GetEnumName(PropertyType, OrdValue);
          end;
        tkLString:
          Result := '''' + GetStrProp(PropOwner, PropertyInfo) + '''';
        tkWString:
          Result := '''' + GetStrProp(PropOwner, PropertyInfo) + '''';
        tkString:
          Result := '''' + GetStrProp(PropOwner, PropertyInfo) + '''';
        tkVariant:
          Result := GetVariantProp(PropOwner, PropertyInfo);
        tkMethod:
          Result := Result + '(' + GetEnumName(TypeInfo(TMethodKind), Ord(GetTypeData(PropertyType)^.MethodKind)) + ')';
      else
        Result := '<' + PropertyType^.Name + '> ('
          + GetEnumName(TypeInfo(TTypeKind), Ord(PropertyKind)) + ')';
      end
    end
  end
  else
    Result := '<unknown>';
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendProperty(const Msg: string; Obj: TObject; const PropName: string);
var
  ComputedMsg: string;
  PropObject: TObject;
  PropValue: string;
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendProperty(Msg, Obj, PropName);
{$ENDIF}
// Functionality is missing
//  wtOverseer:
//    Debugger.LogProperty(Msg, Obj, PropName);
// Simulate the CodeSite method
  else if not Assigned(Obj) then
    Self.DbgSendMsgEx(dbgmProperty, Msg + ' (Object is not assigned)')
  else
  begin
    PropValue := PropertyToString(Obj, PropName, PropObject);
    if Assigned(PropObject) then
      Self.DbgSendObject(Msg, PropObject)
    else
    begin
      ComputedMsg := Msg + ':  ' + PropName + ' = ' + PropValue;
      Self.DbgSendMsgEx(dbgmProperty, ComputedMsg);
    end;
  end;
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Sends a dbgmStream message to the Viewer. The Msg string is displayed in the message
// list, while a description of the data that is transferred when the component referenced
// in the Obj parameter is placed on a stream is displayed in the inspector pane.
//
// If the viewer dosn't support attachments, the attachment can be promoted to
// additional messages by setting the SimulateAttachments property.
//
// ---------------------------------------------------------------------------------------
//

function TAsgDebugMessage.ComponentToStringList(Obj: TComponent): TStringList;
var
  BinaryStream: TMemoryStream;
  TextStream: TMemoryStream;
  TextString: string;
begin
  Result := TStringList.Create;
  BinaryStream := nil;
  TextStream := nil;
  try
// Convert component to binary stream
    BinaryStream := TMemoryStream.Create;
    BinaryStream.WriteComponent(Obj);
    BinaryStream.Position := 0;
// Convert binary representation to text representation
    TextStream := TMemoryStream.Create;
    ObjectBinaryToText(BinaryStream, TextStream);
// Convert text stream to string
    TextStream.Position := 0;
    SetLength(TextString, TextStream.Size);
    TextStream.ReadBuffer(TextString[1], TextStream.Size);
// Split string into lines
    Result.Text := TextString;
  finally
    BinaryStream.Free;
    TextStream.Free;
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendStream(const Msg: string; Obj: TComponent);
var
  List: TStringList;
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendStream(Msg, Obj);
{$ENDIF}
    wtOverseer:
      Debugger.LogComponent(Msg, Obj);
  else
    if not Assigned(obj) then
      Self.DbgSendError(Msg + ' (Component is not assigned)')
    else if SimulateAttachments then
    begin
      List := ComponentToStringList(Obj);
      try
        Self.DbgSendStringListEx(dbgmStream, Msg, List);
      finally
        List.Free;
      end;
    end
    else
      Self.DbgSendMsgEx(dbgmStream, Msg + ' (Attachments not supported)');
  end;
end;

//
// ---------------------------------------------------------------------------------------
//
// Sends a dbgmStringList message to the Viewer. The Msg string is displayed in the message
// list, while the list of string values currently stored in the List parameter is
// displayed in the inspector pane.
//
// If the viewer dosn't support attachments, the stringlist can be promoted to
// additional messages by setting the SimulateAttachments property.
//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendStringList(const Msg: string; List: TStrings);
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendStringList(Msg, List);
{$ENDIF}
    wtOverseer:
      Debugger.LogStringList(Msg, List);
  else
// Use extended version
    Self.DbgSendStringListEx(dbgmStringList, Msg, List);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendStringListEx(MsgType: Integer; const Msg: string; List: TStrings);
var
  Index: Integer;
begin
  if not CheckEnabled then Exit;
  case FViewerType of
{$IFDEF UseCodeSite}
    wtCodeSite:
      CodeSite.SendStringList(Msg, List);
{$ENDIF}
    wtOverseer:
      Debugger.LogStringList(Msg, List);
  else
    begin
      if FSimulateAttachments then
      begin
        Self.DbgSendMsgEx(MsgType, Msg + ' (Attachment added as messages)');
        for Index := 0 to List.Count - 1 do
          Self.DbgSendMsgEx(MsgType, 'Att: ' + List.Strings[Index]);
      end
      else
        Self.DbgSendMsgEx(MsgType, Msg + ' (Attachments not supported)');
    end;
  end;
end;

//
// =======================================================================================
//
// Messages with computed Attachments
//
// =======================================================================================
//
// Sends a dbgmMemoryDump message to the Viewer. The Msg string is displayed in the message
// list, while the dump of the given buffer is displayed in the inspector pane.
//
// If the viewer dosn't support attachments, the stringlist can be promoted to
// additional messages by setting the SimulateAttachments property.
//
// ---------------------------------------------------------------------------------------
//

function TAsgDebugMessage.MemoryToStringList(var Buffer; BufferSize: Integer): TStringList;
type
  TByteArray = array[0..1000000000] of Byte;                // Array size must be greater than the
  TCharArray = array[0..1000000000] of AnsiChar;            // the maximal used BufferSize
var
  OutputLine: string;
  i: Integer;
  k: Integer;
  j: Integer;
  ByteBuffer: TByteArray absolute Buffer;
  CharBuffer: TCharArray absolute Buffer;
begin
  Result := TStringList.Create;
// Header
  Result.Add('');
  OutputLine := Format('%5.4x', [BufferSize]) + ' | ';
  for i := 0 to 15 do
    OutputLine := OutputLine + Format('%3.2x', [i]);
  OutputLine := OutputLine + ' | ..... ANSI .....';
  Result.Add(OutputLine);
  Result.Add(StringOfChar('-', Length(OutputLine)));
// Contents
  i := 0;
  while i < BufferSize do
  begin
// Offset
    k := i;
    OutputLine := Format('%5.4x', [k]) + ' | ';
// Hexdecimal dump
    for j := 0 to 15 do
    begin
      if k < BufferSize then
        OutputLine := OutputLine + Format('%3.2x', [ByteBuffer[k]])
      else
        OutputLine := OutputLine + '   ';
      Inc(k);
    end;
// Separator
    OutputLine := OutputLine + ' | ';
// Character dump
    for j := 0 to 15 do
    begin
      if i < BufferSize then
      begin
        if CharBuffer[i] >= ' ' then
          OutputLine := OutputLine + CharBuffer[i]
        else
          OutputLine := OutputLine + '.';
      end
      else
        OutputLine := OutputLine + ' ';
      Inc(i);
    end;
// Separator
    Result.Add(OutputLine);
  end;
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugMessage.DbgSendMemoryDump(const Msg: string; var Buffer; BufferSize: Integer);
var
  List: TStringList;
begin
  List := MemoryToStringList(Buffer, BufferSize);
  try
    Self.DbgSendStringListEx(dbgmMemoryDump, Msg, List);
  finally
    List.Free;
  end;
end;

//
// =======================================================================================
//

constructor TAsgDebugLevel.Create;
begin
  inherited;
  FRangeFlags := TAABitSet.Create(dbgRangeLimit);
  FRangeNames := TStringList.Create;
end;

//
// ---------------------------------------------------------------------------------------
//

destructor TAsgDebugLevel.Destroy;
begin
  if (Self = DebugLevel) and not DebugMessageClosing then
  begin
    DebugMessage.DbgSendError('Attempt to free the global DebugLevel!');
    Exit;
  end;
  FRangeFlags.Free;
  FRangeFlags := nil;
  FRangeNames.Free;
  FRangeNames := nil;
  inherited;
end;

//
// ---------------------------------------------------------------------------------------
//

function TAsgDebugLevel.GetRangeCount: Integer;
begin
  Result := FRangeNames.Count
end;

//
// ---------------------------------------------------------------------------------------
//

function TAsgDebugLevel.GetRangeFlag(Index: Integer): Boolean;
begin
  Result := FRangeFlags.Bits[Index]
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugLevel.SetRangeFlag(Index: Integer; Value: Boolean);
begin
  FRangeFlags.Bits[Index] := Value;
end;

//
// ---------------------------------------------------------------------------------------
//

function TAsgDebugLevel.GetRangeName(Index: Integer): string;
begin
  Result := FRangeNames.Strings[Index];
end;

//
// ---------------------------------------------------------------------------------------
//

procedure TAsgDebugLevel.SetRangeName(Index: Integer; Value: string);
begin
  FRangeNames.Strings[Index] := Value;
end;

//
// ---------------------------------------------------------------------------------------
//

function TAsgDebugLevel.AddRange(const RangeName: string): Integer;
begin
  Result := FRangeNames.IndexOf(RangeName);
  if Result = -1 then
    Result := FRangeNames.Add(RangeName);
end;

//
// ---------------------------------------------------------------------------------------
//

function TAsgDebugLevel.FindRange(const RangeName: string): Integer;
begin
  Result := FRangeNames.IndexOf(RangeName);
end;

//
// =======================================================================================
//

initialization
  DebugMessageClosing := True;
  DebugMessage := TAsgDebugMessage.Create;
  Debuglevel := TAsgDebugLevel.Create;
  DebugMessageClosing := False;
finalization
  DebugMessageClosing := True;
  DebugMessage.Free;
  DebugMessage := nil;
  DebugLevel.Free;
  DebugLevel := nil;
end.

